import copy
from collections import Sequence

from pds4_logging import logger_init

# Safe import of OrderedDict
try:
    from collections import OrderedDict
except ImportError:
    from ordered_dict_26 import OrderedDict

# Initialize the logger
logger = logger_init()

#################################

class ObjectList(Sequence):
    """ ObjectList class; stores label and all objects of a PDS4 observational product

    Allows both Dict-like and List-like access. This class is returned by read_pds4().
    """

    def __init__(self):
        super(ObjectList, self).__init__()

        self._objects = []

        self._label = None
        self._read_in_log = None

    def __getitem__(self, key):

        if isinstance(key, (int, long)) or isinstance(key, slice):
            object = self._objects[key]

        else:

            # Reuse tuple-search logic for simple (ie, str) name and lid searches
            if not isinstance(key, tuple):
                key = (key,) + (0,)

            # Search by local identifier
            object = self._get_object_by_id(key, 'local_identifier')

            # Search by name
            if object is None:
                object = self._get_object_by_id(key, 'name')

        return object

    def __len__(self):
        return len(self._objects)

    @property
    def label(self):
        return self._label

    @property
    def data(self):

        data_list = [obj.data for obj in self._objects]
        return data_list

    @property
    def read_in_log(self):
        return self._read_in_log

    def append_object(self, object):
        self._objects.append(object)

    def set_label(self, label):
        self._label = label

    def set_log(self, log):
        self._read_in_log = log

    def _get_object_by_id(self, key, id_type):

        object = None

        object_lids = [obj.meta_data.get(id_type) for obj in self._objects]
        object_match_idx = [i for i, lid in enumerate(object_lids) if lid == key[0]]

        if len(object_match_idx) > key[1]:
            object = self._objects[object_match_idx[key[1]]]

        return object


class Object(object):
    """ Object class; stores meta data about a single PDS4 data object, the object's label and its data """

    def __init__(self, parent_filename, object_label, object_data=None):
        super(Object, self).__init__()

        self.id = None
        self.type = object_label.tag
        self.parent_filename = parent_filename

        self.label = object_label
        self.data = object_data

        self._meta_data = OrderedDict()

    @property
    def meta_data(self):
        return self._meta_data


class Data(OrderedDict):
    """ Data class; stores data read from file

    Allows both Dict-like and List-like access. This object can contain nested Data objects, such as for group
    repetitions. The meta_data(value) stores information about the value (e.g. if value is a table field, then meta_data
    would be a Field object containing things like field format).
    """

    def __init__(self):

        super(Data, self).__init__()
        self._meta_data = OrderedDict()

    def __getitem__(self, key):

        if isinstance(key, (int, long)) or isinstance(key, slice):
            item = self[self.keys()[key]]
        else:
            item = super(Data, self).__getitem__(key)

        return item

    def __setitem__(self, key, value, dict_setitem=dict.__setitem__):
        raise TypeError('Data object does not support item assignment, use .set() method')

    # This method overwrites copy.deepcopy(Data) so that it works without allowing Data to be
    # initialized with *args and **kwd, which would bypass the requirement to specify meta_data for each key. Note that
    # for anything but the smallest data this method will be extremely expensive unless use_numpy is on
    def __deepcopy__(self, memo):

        deepcopied_data = Data()

        # Recursively deepcopy all values
        for key, value in self.iteritems():
            deepcopied_data.set(key, copy.deepcopy(value, memo), copy.deepcopy(self.meta_data(key), memo))

        return deepcopied_data

    def set(self, key, data, meta_data):

        if isinstance(key, (int, long)):
            raise NameError('Data object does not support integer keys')

        self._meta_data[key] = meta_data
        super(Data, self).__setitem__(key, data)

    def meta_data(self, key):

        if isinstance(key, (int, long)):
            meta_data = self._meta_data[self._meta_data.keys()[key]]
        else:
            meta_data = self._meta_data[key]

        return meta_data

    # Returns an equivalent, uncopied, list (stripping the meta_data)
    def to_list(self):

        list_dt = list()

        # Recursively convert all nested Data objects to List
        for key, value in self.iteritems():

            if isinstance(value, Data):
                list_dt.append(value.to_list())
            else:
                list_dt.append(value)

        return list_dt

    # Returns an equivalent, uncopied, OrderedDict (stripping the meta_data)
    def to_ordered_dict(self):

        ordered_dict = OrderedDict(self)

        # Recursively convert all nested Data objects to OrderedDicts
        for key, value in ordered_dict.iteritems():

            if isinstance(value, Data):
                ordered_dict[key] = value.to_ordered_dict()

        return ordered_dict


class DataObject(object):
    """ Stores meta data about a single Data Object (e.g. Table_Character,  Array_2D_Image,  Array_3D, etc) """

    def __init__(self, name, type, parent_filename, parent_label):

        self.name = name
        self.type = type
        self.parent_filename = parent_filename
        self.parent_label = parent_label