import sys
import logging


# Initializes the logger and its handlers
def logger_init():

    logging.setLoggerClass(PDS4Logger)
    logger = logging.getLogger('PDS4Logger')
    logger.setLevel(logging.DEBUG)

    # If this is a new logger then initialize its config
    if not logger.handlers:

        # Create the user handler. This handler outputs to stdout and can be silenced by user
        # (via use of quiet or --quiet).
        user_handler = PDS4StreamHandler('user_handler')

        # Create the log handler. This handler does not output to screen and should not be silenced
        app_handler = PDS4SilentHandler('log_handler')

        # Create the formatter and add it to the handlers
        formatter = PDS4Formatter()
        user_handler.setFormatter(formatter)
        app_handler.setFormatter(formatter)

        # Add handlers to logger. Must be added in order, first user_handler, then app_handler.
        logger.addHandler(user_handler)
        logger.addHandler(app_handler)

    return logger


class PDS4Logger(logging.Logger):
    """ Customer Logger that supports getting handlers by name and has a quiet() method """

    def get_handler(self, handler_name):

        for handler in self.handlers:

            if handler.name == handler_name:
                return handler

        return None

    def quiet(self, handler_name='user_handler'):
        self.get_handler(handler_name).setLevel(logging.ERROR)

class PDS4StreamHandler(logging.StreamHandler):
    """ Custom StreamHandler that a name """

    def __init__(self, name):

        # Using try due to stream parameter being renamed in Python <2.7)
        try:
            logging.StreamHandler.__init__(self, stream=sys.stdout)
        except TypeError:
            logging.StreamHandler.__init__(self, strm=sys.stdout)

        self._name = name

    @property
    def name(self):
        return self._name


class PDS4SilentHandler(PDS4StreamHandler):
    """ Custom StreamHandler that saves emitted records to self.records  """

    def __init__(self, name):

        PDS4StreamHandler.__init__(self, name)

        self.records = []

    def emit(self, record):
        self.records.append(record)

    def to_string(self, start_i=0, end_i=None):

        string_message = ''

        for record in self.records[start_i:end_i]:
            string_message += self.format(record) + "\n"

        return string_message


class PDS4Formatter(logging.Formatter):
    """ Custom formatter that varies format according to log level """

    def format(self, record):

        formatter = logging.Formatter('%(message)s')
        formatted_value = logging.Formatter.format(formatter, record)

        if record.levelno != logging.INFO:
            formatted_value = record.levelname.capitalize() + ': ' + formatted_value

        return formatted_value