import os

import pds4_read_tables
from pds4_general_objects import ObjectList
from pds4_read_label import read_label
from pds4_logging import logger_init

# Initialize the logger
logger = logger_init()

#################################


def pds4_read(filename, quiet=False, use_numpy=False, object_num=None, object_name=None, object_lid=None):

    """ Reads PDS4 compliant data into a Data object

        NOTES:

            Currently supports only the objects Table_Character and Table_Binary. Packed bit fields in Table_Binary are
            not yet supported.

        Parameters:

            filename: str
                The filename, including full or relative path if necessary, of the PDS4 label describing the data
            quiet: bool
                Suppresses all info/warnings from being output
            use_numpy: bool
                Returned data will be a NumPy array and use NumPy data types
            object_num: str
                If given then only reads the data object specified by zero-based order (integer)
            object_name: str
                If given then only reads the data object specified by name
            object_lid: str
                If given then only reads the data object specified by local identifier

        Returns:

            ObjectList
                Contains PDS4 object and label data. Can be treated/accessed/used like a Dict or List.

        Example usage:

            obj_list = pds4_read('Example_Label.xml')

            Example Label:

                Table: unnamed
                    Field: order
                    Field: wavelength - (1)
                Table: Observations
                    Field: order
                    Group: unnamed; repetitions: 2
                        Field: pos_vector - (2)

            Accessing Example Label data:

                You may use any combination of Dict or List like indicies.

                unnamed_table = obj_list[0]
                             or obj_list['TABLE_0']

                obs_table = obj_list[1]
                         or obj_list['Observations']

                (1) unnamed_table.data['wavelength']
                 or unnamed_table.data[1]

                (2) obs_table.data['GROUP_0'][0]['pos_vector']
                    obs_table.data['GROUP_0'][1]['pos_vector']

            Accessing Example Label meta data:

                You may use any combination of Dict or List like indicies.

                (1) unnamed_table.data.meta_data('wavelength')['unit']
                 or unnamed_table.data.meta_data(1)['unit']

                (2) obs_table.data['GROUP_0'][0].meta_data('pos_vector')['unit']
                    obs_table.data['GROUP_0'][1].meta_data('pos_vector')['unit']

            Accessing Example Label:

                The returned value is an ElementTree object.

                Entire label:
                    obj_list.label

                Part of label describing Observations table:
                    obj_list['Observations'].label
                    obj_list[1].label

    """

    # Set logger to only emit error messages if requested
    if quiet:
        logger.quiet()

    # Initialize the log recording
    record_start = len(logger.get_handler('log_handler').records)

    # Read-in the PDS4 label
    label_xml_root = read_label(filename)

    # Extract directory of label
    label_dir = os.path.dirname(filename)

    # Variable that stores byte data for each data file_xml referenced in the label
    byte_data = []

    # Loop over each File Area Observation and read-in all data files
    for i, file_xml in enumerate(label_xml_root.findall('.//File_Area_Observational')):

        data_filename = file_xml.find('.//file_name').text
        data_filepath = os.path.join(label_dir, data_filename)

        logger.info('Now reading data file_xml: ' + data_filename)

        # Read in the entire data file_xml in binary form into byte_data
        try:
            file_handler = open(data_filepath, 'rb')
            file_byte_data = file_handler.read()
            file_handler.close()

            byte_data.append(file_byte_data)
        except IOError:
            raise IOError("Unable to read data file '" + data_filepath + "' found in label")

    # Convert byte_data to tuple (in order to a pass a non-mutable variable as argument into read_table and read_array)
    byte_data = tuple(byte_data)

    # Variable that stores all returned data objects
    object_list = ObjectList()
    object_list.set_label(label_xml_root)

    table_objects = pds4_read_tables.read_tables(label_xml_root, byte_data, use_numpy=use_numpy, quiet=quiet,
                                                 object_num=object_num, object_name=object_name, object_lid=object_lid)

    for object in table_objects:
        object_list.append_object(object)

    # Save the log recording
    object_list.set_log(logger.get_handler('log_handler').to_string(record_start))

    if len(object_list) == 0:
        raise ValueError('No viable PDS4 data objects found')

    return object_list


if __name__ == '__main__':

    # Safe import of argparse
    try:
        import argparse
    except ImportError:
        import pds4_tools.argparse_26 as argparse

    # Create program arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("filename", help="Filename, including full path, of the label")
    parser.add_argument("--use_numpy", help="Returned data will be a numpy array and use numpy data types",
                        action="store_true", default=False)
    parser.add_argument("--quiet", help="Suppresses all info/warnings", action="store_true", default=False)
    parser.add_argument("--object_num", help="Only reads the data object specified by zero-based order (integer)")
    parser.add_argument("--object_name", help="Only reads the data object specified by name")
    parser.add_argument("--object_lid", help="Only reads the data object specified by local identifier")

    args = parser.parse_args()

    pds4_read(args.filename, use_numpy=args.use_numpy, quiet=args.quiet, object_num=args.object_num,
              object_name=args.object_name, object_lid=args.object_lid)