import copy
from math import log
import xml.etree.ElementTree as ET

import pds4_read_tables_ascii
import pds4_read_tables_binary
from pds4_general_objects import ObjectList, Data
from pds4_table_objects import TableStructure, TableObject, FieldUniformlySampled, is_group
from pds4_helpers import is_int
from pds4_logging import logger_init

# Safe import of numpy (not required)
try:
    import numpy as np
except ImportError:
    pass

# Initialize the logger
logger = logger_init()

#################################

# Creates the storage variable, including all stored data, for Uniformly_Sampled
def _make_uniformly_sampled(table_xml, use_numpy=False):

    uniformly_sampled_xml = table_xml.find('Uniformly_Sampled')
    uniformly_sampled_data = []
    uniformly_sampled_dict = {}

    if uniformly_sampled_xml is not None:

        uni_sampled = FieldUniformlySampled()
        uni_sampled.load(uniformly_sampled_xml)

        if ('scale' not in uni_sampled) or (uni_sampled['scale'] == 'Linear'):
            current_value = uni_sampled['first_value']

            while current_value <= uni_sampled['last_value']:
                uniformly_sampled_data.append(current_value)
                current_value += uni_sampled['interval']

        elif uni_sampled['scale'] == 'Logarithmic':

            x = log(uni_sampled['first_value']) / log(10)
            current_value = uni_sampled['first_value']
            counter = 0

            while current_value < uni_sampled['last_value']:
                current_value = 10 ** (uni_sampled['interval'] * counter + x)
                counter += 1

                uniformly_sampled_data.append(current_value)

        elif uni_sampled['scale'] == 'Exponential':

            x = 10. ** uni_sampled['first_value']
            current_value = uni_sampled['first_value']
            counter = 0

            while current_value < uni_sampled['last_value']:
                current_value = log(uni_sampled['interval'] * counter + x) / log(10)
                counter += 1

                uniformly_sampled_data.append(current_value)

        # Return data values to int if this was intended
        if is_int(uni_sampled['interval']) and is_int(uni_sampled['first_value']) and is_int(uni_sampled['last_value']):

            all_ints = True

            for value in uniformly_sampled_data:
                if not value.is_integer():
                    all_ints = False

            if all_ints:
                uniformly_sampled_data = [int(value) for value in uniformly_sampled_data]

        if use_numpy:
            uniformly_sampled_dict = {uni_sampled['name']: np.asarray(uniformly_sampled_data)}
        else:
            uniformly_sampled_dict = {uni_sampled['name']: tuple(uniformly_sampled_data)}

    return uniformly_sampled_dict


# Extracts and properly formats the data for a single table, as specified by table_object, from byte_data
def _extract_table_data(file_byte_data, table_object, table_structure, use_numpy=False):

    # Variable that will store final version of all returned data from this table
    data = Data()

    # Add Uniformly_Sampled to data if it exists
    uniformly_sampled = _make_uniformly_sampled(table_object.label, use_numpy)
    if uniformly_sampled:
        data.set(uniformly_sampled.keys()[0], uniformly_sampled.values()[0], None)

    # This for loop creates the 'data' variable, which will store all columns and their values for the table that was
    # just read in. Begin by looping over each column
    for i in range(0, table_structure.num_items):

        # Skip groups; we need the i to be the index in table_structure below but do not need groups
        if is_group(table_structure[i]):
            continue

        field = table_structure[i]
        extracted_bytes = []

        # Pre-extract necessary variables to speed up computation time
        offset = table_object.meta_data['offset']
        field_location = field['location'] - 1
        record_length = table_object.record['record_length']
        field_length = field['length']

        # Extract the byte data for this specific column from byte_data
        for a in range(0, table_object.meta_data['records']):
            start_idx = offset + field_location + (a * record_length)
            end_idx = start_idx + field_length

            extracted_bytes.append(file_byte_data[start_idx:end_idx])

        # Cast the byte data into appropriate data types
        if table_object.type == 'Table_Character':
            extracted_data = pds4_read_tables_ascii.data_type_convert(field['data_type'], extracted_bytes, use_numpy)

        elif table_object.type == 'Table_Binary':
            extracted_data = pds4_read_tables_binary.data_type_convert(field['data_type'], extracted_bytes, use_numpy)

        # Make scaling_factor and value_offset adjustments
        if ('scaling_factor' in field) or ('value_offset' in field):

            for j, datum in enumerate(extracted_data):
                if 'scaling_factor' in field:
                    extracted_data[j] *= field['scaling_factor']

                if 'value_offset' in field:
                    extracted_data[j] += field['value_offset']

        # Retrieve indicies of all the parent groups of this column (needed to extract the group names for the 'data'
        # variable)
        parent_group_idxs = []

        for j in range(0, field.group_level):

            if parent_group_idxs:
                parent_idx = table_structure.get_parent_by_idx(parent_group_idxs[j - 1], return_idx=True)
            else:
                parent_idx = table_structure.get_parent_by_idx(i, return_idx=True)

            parent_group_idxs.append(parent_idx)

        # Find the appropriate place in the 'data' variable to store the newly extracted column by going deeper into
        # the 'data' variable (e.g. from data['GROUP_0'][0] to data['GROUP_0'][0]['GROUP_0'][0]) with each iteration
        # of this loop
        data_ref = data

        for j in range(len(parent_group_idxs) - 1, -1, -1):
            group = table_structure[parent_group_idxs[j]]

            # Find Groups with identical names and in identical group locations inside Data (without special
            # handling these groups would overwrite each other)
            if j == len(parent_group_idxs) - 1:
                parent_group_idxs.append(-1)

            # Current field has a parent group, and that parent group has its own parent group (if j condition above is
            # satisfied then this would be the entire record). In the first iteration of the j loop, the below variable
            # contains the children of that last parent group (one such child is the parent group of the current field)
            children_of_grandparent_group = table_structure.get_children_by_idx(parent_group_idxs[j+1])

            # identical_group_idxs will always return at least one match (the direct parent group of current field)
            identical_group_idxs = [1 + parent_idx + parent_group_idxs[j + 1]
                                    for parent_idx, grandparent_child in enumerate(children_of_grandparent_group)
                                    if grandparent_child['name'] == group['name']
                                    if grandparent_child.group_level == group.group_level
                                    if grandparent_child.group_repetition == group.group_repetition
                                    if is_group(grandparent_child)]

            # Create the group if it does not exist already
            if group['name'] not in data_ref:

                dt_tuple = tuple([Data() for k in range(0, group['repetitions'])])

                # Handle special case of Groups with identical names and in identical locations
                if len(identical_group_idxs) >= 2:
                    dt_tuple_repeat = tuple([copy.deepcopy(dt_tuple)
                                             for k in range(0, len(identical_group_idxs))])

                    group_repeat = tuple(table_structure[identical_group_idxs[k]]
                                         for k in range(0, len(identical_group_idxs)))

                    data_ref.set(group['name'], dt_tuple_repeat, group_repeat)

                # Normal group creation
                else:
                    data_ref.set(group['name'], dt_tuple, group)

            # Access the group
            data_ref = data_ref[group['name']]

            # Access the correct group in special case of Groups with identical names and in identical locations
            if len(identical_group_idxs) >= 2:
                data_ref = data_ref[identical_group_idxs.index(parent_group_idxs[0])]

            # Get group_repetition through Field once already inside the Field's direct parent Group
            if j == 0:
                data_ref = data_ref[field.group_repetition]

            # Get group_repetition from next parent Group otherwise
            else:
                next_group = table_structure[parent_group_idxs[j-1]]
                data_ref = data_ref[next_group.group_repetition]

        # Insert the newly extracted column into the 'data' variable
        if not use_numpy:
            extracted_data = tuple(extracted_data)

        # Correctly append this field
        if field['name'] not in data_ref:
            data_ref.set(field['name'], extracted_data, field)

        # Correctly append this field if there is another field with this name at this location (without special
        # handling these fields would overwrite each other)
        else:
            
            field_name = field['name']

            # Append to Field that is currently a nested tuple and thus there was another field with this name before
            if hasattr(data_ref[field_name][0], '__len__') and (not isinstance(data_ref[field_name][0], str)):

                field_metadata = (data_ref.meta_data(field_name)) + (field,)
                if not use_numpy:
                    data_ref.set(field_name, (data_ref[field_name]) + (extracted_data,), field_metadata)

            # Append to Field that is currently tuple and thus until now it was the only field with this name
            else:

                field_metadata = (data_ref.meta_data(field_name),) + (field,)
                if not use_numpy:
                    data_ref.set(field_name, (data_ref[field_name],) + (extracted_data,), field_metadata)

            if use_numpy:
                data_ref.set(field_name, np.vstack((data_ref[field_name], extracted_data)), field_metadata)

    return data


def read_tables(label_xml_root, file_byte_data, quiet=False, use_numpy=False, object_num=None, object_name=None,
                object_lid=None):

    # Set logger to only emit error messages if requested
    if quiet:
        logger.quiet()

    # Storage variable for all PDS4 objects
    object_list = ObjectList()
    object_list.set_label(label_xml_root)

    table_types = ['Table_Character', 'Table_Binary']

    # Loop over each File Area Observation
    for i, file_xml in enumerate(label_xml_root.findall('.//File_Area_Observational')):

        data_filename = file_xml.find('.//file_name').text

        # Work around ElementTree not having ability to findall based on multiple conditions and preserving order
        tables_xml = [element for element in file_xml if element.tag in table_types]

        # Loop over each Table inside the File Area
        for j, table_xml in enumerate(tables_xml):

            # Set a local_identifier for the table of it has no local identifier or name in the label
            if (table_xml.find('name') is None) and (table_xml.find('local_identifier') is None):

                lid = ET.fromstring('<local_identifier>TABLE_{0}</local_identifier>'.format(j))
                table_xml.insert(0, lid)

            # Create the data object for this table
            table_object = TableObject(data_filename, table_xml)

            # Determine if this table should be saved
            if object_num is not None and object_num != j:
                continue

            if object_name is not None and table_object.meta_data['name'] != object_name:
                continue

            if object_lid is not None and table_object.meta_data['local_identifier'] != object_lid:
                continue

            logger.info('Now processing a {0} object: {1}'.format(table_object.type, table_object.id))

            # Stores data about the table's groups and fields
            table_structure = TableStructure()
            table_structure.load(table_xml)

            # Read the table from file into a variable
            table_object.data = _extract_table_data(file_byte_data[i], table_object, table_structure, use_numpy)

            # Save the read-in data object
            object_list.append_object(table_object)

    return object_list
