import copy
import re
from collections import Sequence

from pds4_general_objects import Object
from pds4_helpers import cast_int_or_float, xml_to_dict
from pds4_logging import logger_init

# Safe import of OrderedDict

try:
    from collections import OrderedDict
except ImportError:
    from ordered_dict_26 import OrderedDict

# Initialize the logger
logger = logger_init()

#################################


class TableStructure(Sequence):
    """ Stores meta data about a single table's groups and fields

    Once .load() is used, this object will read the XML Record for a single table and create a representation
    of the table, ordered by byte location, by adding the appropriate Field and Group objects.
    """

    def __init__(self):

        self._struct = []
        self._num_items = 0

        self._table_type = None

        super(TableStructure, self).__init__()

    def __getitem__(self, key):
        return self._struct[key]

    def __len__(self):
        return self._num_items

    # Fully initializes the TableStructure with all appropriate Field and Group objects
    def load(self, table_xml):

        self._table_type = table_xml.tag.split('_')[1]
        record_character_xml = table_xml.find('Record_' + self._table_type)

        # Add all fields and groups to the TableStructure
        self._add_fields_and_groups(record_character_xml)

        # Set group names for all groups, and full names for all fields and groups (done prior to cloning, such that
        # cloned fields and groups preserve the names)
        self._create_group_names()
        self._create_full_names()

        # Create a pre-cloned copy, used for validation
        pre_cloned_table_struct = copy.deepcopy(self)
        pre_cloned_table_struct.adjust_to_absolute_location()

        # Clone repeating group fields and adjust locations to their absolute values instead of the default
        # relative values
        self._clone_all_groups()
        self.adjust_to_absolute_location()

        # Set full names for all fields and groups, including repetitions this time
        self._create_full_names(include_repetitions=True)

        # Perform basic sanity and basic validation checking on the table
        self._validate_table(pre_cloned_table_struct, record_character_xml)
        self._validate_table_sbn(pre_cloned_table_struct)

    @property
    def struct(self):
        return self._struct

    @property
    def num_items(self):
        return self._num_items

    # Returns a new TableStructure that contains the children of the idx element. Returns only the immediate children
    # if direct_only is set to True
    def get_children_by_idx(self, parent_idx, direct_only=False):

        child_table_structure = TableStructure()

        if parent_idx == -1:
            parent_element_group_lvl = 0

        else:
            parent_element_group_lvl = self._struct[parent_idx].group_level + 1

        for i in range(parent_idx + 1, self._num_items):
            cur_element = self._struct[i]

            if cur_element.group_level >= parent_element_group_lvl:

                if cur_element.group_level == parent_element_group_lvl or not direct_only:
                    child_table_structure._append(cur_element)

            else:
                break

        return child_table_structure

    # Returns the parent group of the child given by child_idx. Returns the index of the parent group if  return_idx
    # is set to True.
    def get_parent_by_idx(self, child_idx, return_idx=False):
        child_element = self._struct[child_idx]
        child_group_level = child_element.group_level

        for i in range(child_idx - 1, -1, -1):
            cur_element = self._struct[i]

            if (cur_element.group_level + 1) == child_group_level and is_group(cur_element):

                if return_idx:
                    return i
                else:
                    return cur_element

        return -1

    # Adjusts group_location and field_location in TableStructure to be the absolute locations from the beginning of
    # the data file instead of the relative location used by PDS4 schemas inside <Group_Field_*>
    def adjust_to_absolute_location(self):
        group_indicies = [i for i, item in enumerate(self._struct)
                          if is_group(item)
                          if item.group_level == 0]

        for group_idx in group_indicies:
            child_elements = self.get_children_by_idx(group_idx).struct

            for i, element in enumerate(child_elements):

                parent_idx = self.get_parent_by_idx(group_idx + i + 1, return_idx=True)

                if parent_idx != -1:
                    element['location'] += self._struct[parent_idx]['location'] - 1

    def _insert(self, key, field_or_group):
        self._struct.insert(key, field_or_group)
        self._num_items += 1

    def _append(self, field_or_group):
        self._struct.append(field_or_group)
        self._num_items += 1

    # Adds all <Field_*>, <Group_Field_*>, and all sub-<Field_*> and sub-<Group_Field_*>, to TableStructure which are
    # direct children of xml_parent
    def _add_fields_and_groups(self, xml_parent, group_level=0):

        field_tag = 'Field_' + self._table_type
        group_tag = 'Group_Field_' + self._table_type

        # Work around ElementTree not having ability to findall based on multiple conditions and preserving order
        elements_xml = [element for element in xml_parent
                        if element.tag == field_tag or
                        element.tag == group_tag]

        # Iterate over all fields and groups
        for element in elements_xml:

            # Append Fields to Data
            if element.tag == field_tag:

                if self._table_type == 'Character':
                    appended_field = FieldCharacter()
                elif self._table_type == 'Binary':
                    appended_field = FieldBinary()
                elif self._table_type == 'Delimited':
                    appended_field = FieldDelimited()
                else:
                    raise ValueError('Unknown table type: ' + self._table_type)

                appended_field.load(element)
                self._append(appended_field)

                appended_field.group_level = group_level

            # Append Groups (and recursively sub-Fields and sub-Groups) to Data
            else:

                appended_group = Group()
                appended_group.load(element)
                self._append(appended_group)

                appended_group.group_level = group_level
                self._add_fields_and_groups(element, group_level + 1)

    # Clones/repeats all <Group_Field_*> in TableStructure according to the 'repetition' field
    def _clone_all_groups(self):
        index_adjust = 0
        group_indicies = [i for i, item in enumerate(self._struct)
                          if is_group(item)
                          if not item.cloned]

        if not group_indicies:
            return

        for group_idx in group_indicies:
            start_length = self._num_items
            self._clone_group(group_idx + index_adjust)
            index_adjust += self._num_items - start_length

        self._clone_all_groups()

    # Clones a single <Group_Field_*>, adjusting location to account for the repetition
    def _clone_group(self, idx):
        group = self._struct[idx]

        cloned_group = copy.deepcopy(self.get_children_by_idx(idx))
        current_idx = idx + cloned_group.num_items

        for i in range(1, group['repetitions']):

            # Insert each child of the group being cloned individually
            for item in copy.deepcopy(cloned_group):

                current_idx += 1
                self._insert(current_idx, item)
                parent_idx = self.get_parent_by_idx(current_idx, return_idx=True)

                # If element is a direct child of the group being cloned then adjust its location and group_repetition
                if parent_idx == idx:
                    group_length = self._struct[parent_idx]['length']
                    group_repetitions = self._struct[parent_idx]['repetitions']

                    item.group_repetition = i
                    item['location'] += group_length / group_repetitions * i

        group.cloned = True

    # Recursively creates names for groups if they do not exist in the label
    def _create_group_names(self, group_level=0):

        group_indicies = [i for i, item in enumerate(self._struct)
                          if is_group(item)
                          if item.group_level == group_level]

        if not group_indicies:
            return

        prev_parent = -2
        counter = 0

        for group_idx in group_indicies:

            group = self._struct[group_idx]
            parent_idx = self.get_parent_by_idx(group_idx, return_idx=True)

            if parent_idx == prev_parent:
                counter += 1
            else:
                counter = 0

            if 'name' not in group:
                group['name'] = 'GROUP_' + str(counter)

            group.full_name = group['name']
            prev_parent = parent_idx

        self._create_group_names(group_level + 1)

    # Creates the full names (ie, name preceded by all parent group names) for all fields and groups. Setting
    # include_repetitions to True will have the group repetition following each group name
    def _create_full_names(self, include_repetitions=False):

        for i, item in enumerate(self._struct):

            # Begin by setting full_name to the item name
            full_name = item['name']
            idx = i

            parent_idx = self.get_parent_by_idx(idx, return_idx=True)
            grandparent_children = self.get_children_by_idx(parent_idx)

            # Find _struct indicies of items having the same name and group location as current item
            if is_group(item):
                identical_idxs = [1 + child_idx + parent_idx
                                  for child_idx, grandparent_child in enumerate(grandparent_children)
                                  if grandparent_child['name'] == item['name']
                                  if grandparent_child.group_level == item.group_level
                                  if grandparent_child.group_repetition == item.group_repetition
                                  if is_group(grandparent_child)]

            else:
                identical_idxs = [1 + child_idx + parent_idx
                                  for child_idx, grandparent_child in enumerate(grandparent_children)
                                  if grandparent_child['name'] == item['name']
                                  if grandparent_child.group_level == item.group_level
                                  if grandparent_child.group_repetition == item.group_repetition
                                  if not is_group(grandparent_child)]

            # Add count for groups and fields that have the same name and same group location (otherwise these groups
            # and fields would have exactly the same name)
            if len(identical_idxs) >= 2:
                full_name += ' [{0}]'.format(identical_idxs.index(i))

            # Add field parent group's full name (this should already contain names of all prior groups)
            if parent_idx != -1:
                group = self._struct[parent_idx]

                if include_repetitions:

                    group_repetition = item.group_repetition
                    full_name = '{0} (Rep {1}): {2}'.format(group.full_name, group_repetition, full_name)

                else:
                    full_name = '{0}: {1}'.format(group.full_name, full_name)

            item.full_name = full_name

    # Performs basic sanity checks on the table and basic verification of some PDS4 standards not tested by schema
    def _validate_table(self, pre_cloned_table_struct, record_character_xml):

        # Check that number of fields and groups claimed matches number of <Field_*> and <Group_Field_*>
        num_rec_fields = int(record_character_xml.find('fields').text)
        num_rec_groups = int(record_character_xml.find('groups').text)

        groups = [item for item in pre_cloned_table_struct if is_group(item) and item.group_level == 0]
        fields = [item for item in pre_cloned_table_struct if not is_group(item) and item.group_level == 0]

        fields_mismatch_warn = '<fields> value does not match number of <Field_{0}> in'.format(self._table_type)
        groups_mismatch_warn = '<groups> value does not match number of <Group_Field_{0}> in'.format(self._table_type)

        if len(fields) != num_rec_fields:
            logger.warning('{0} Record_{1}.'.format(fields_mismatch_warn, self._table_type))

        if len(groups) != num_rec_groups:
            logger.warning('{0} Record_{1}.'.format(groups_mismatch_warn, self._table_type))

        for i, item in enumerate(pre_cloned_table_struct):

            full_location = item.full_name.replace(': ', ' -> ')
            full_location_warn = "'{0}' (full location: '{1}')".format(item['name'], full_location)

            if is_group(item):

                # Check number of fields and groups matching in nested groups
                children = pre_cloned_table_struct.get_children_by_idx(i, direct_only=True)

                groups = [child for child in children if is_group(child)]
                fields = [child for child in children if not is_group(child)]

                if len(fields) != item['fields']:
                    logger.warning(fields_mismatch_warn + ' group ' + full_location_warn)

                if len(groups) != item['groups']:
                    logger.warning(groups_mismatch_warn + ' group ' + full_location_warn)

                # Check that group_length is evenly divisible by group_repetitions
                if item['length'] % item['repetitions'] != 0:
                    raise Exception("Group length '{0}' must be evenly divisible by the number of repetitions '{1}' "
                                    "for group {2}".format(item['length'], item['repetitions'], full_location_warn))

            elif 'format' in item:

                # Ensure field_format requirements
                valid_field_format = re.compile('%([\+,-])?([0-9]+)(\.([0-9]+))?([doxfes])')

                if not valid_field_format.match(item['format']):
                    logger.warning("field_format '{0}' does not conform to standards for field {1} "
                                   .format(item['format'], full_location_warn))

                else:

                    is_numeric = {'ASCII_Real': True,
                                  'ASCII_Integer': True,
                                  'ASCII_NonNegative_Integer': True,
                                  'ASCII_Numeric_Base2': True,
                                  'ASCII_Numeric_Base8': True,
                                  'ASCII_Numeric_Base16': True}.get(item['data_type'], False)

                    sub_exp = valid_field_format.search(item['format']).groups()

                    plus_minus = sub_exp[0]
                    width = sub_exp[1]
                    precision = sub_exp[3]
                    specifier = sub_exp[4]

                    if plus_minus is not None:

                        if plus_minus == '+' and not is_numeric:
                            logger.warning("field_format '{0}' may not have '+' for a numeric data_type '{1}' for "
                                           "field {2}".format(item['format'], item['data_type'], full_location_warn))

                        elif plus_minus == '-' and is_numeric:
                            logger.warning("field_format '{0}' may not have '-' for a non-numeric data_type '{1}' for "
                                           "field {2}".format(item['format'], item['data_type'], full_location_warn))

                    if int(width) != item['length'] and self._table_type == 'Character':
                        logger.warning("in Table_Characters field_format '{0}' width must be equal to field_length "
                                       "'{1}' for field {2}".format(item['format'], item['length'], full_location_warn))

                    if precision is not None and specifier == 's' and int(width) != int(precision):
                        logger.warning("field_format '{0}' precision must be equal to width '{1}' for a string format "
                                       "for field {2}".format(item['format'], width, full_location_warn))

    # Additional sanity checks on TableStructure, looking for invalid content according to PDS-SBN standards but
    # valid PDS4
    def _validate_table_sbn(self, pre_cloned_table_struct):

        groups = [item for item in pre_cloned_table_struct if is_group(item)]

        if groups:
            logger.warning('-SBN-: Group fields are used by this table')

        for i, item in enumerate(pre_cloned_table_struct):

            full_location = item.full_name.replace(': ', ' -> ')
            full_location_warn = "'{0}' (full location: '{1}')".format(item['name'], full_location)

            if is_group(item):

                if item['repetitions'] < 2:
                    logger.warning("-SBN-: A group's repetitions should be at least 2 for Group " + full_location_warn)

            else:

                # Check that expected elements for each Field exist and conform to SBN standards

                if (not groups) and ('number' not in item):
                    logger.warning('-SBN-: a field_number is expected for all Fields in flat tables, but one was not '
                                   'found for Field ' + full_location_warn)

                if 'format' not in item:
                    logger.warning('-SBN-: a field_format is required for all Fields, but was not found for Field '
                                   + full_location_warn)

                if 'description' not in item:
                    logger.warning('-SBN-: description is expected for all Fields but was not found for Field '
                                   + full_location_warn)

                if ('unit' in item) and (item['unit'] == 'N/A' or item['unit'].strip() == ''):
                    logger.warning('-SBN-: blank or N/A should not be used for units but was found for Field '
                                   + full_location_warn)

        # Check for overlapping Field locations
        fields = [item for item in pre_cloned_table_struct if not is_group(item)]

        if len(fields) > 1000:
            logger.warning('-SBN-: Large number of Fields, field_location overlap checking skipped.')
            return

        for i, item in enumerate(fields):

            for j, inner_item in enumerate(fields):

                if i == j:
                    continue

                item_end_location = item['location'] + item['length'] - 1

                if item['location'] <= inner_item['location'] <= item_end_location:

                    full_location = item.full_name.replace(': ', ' -> ')
                    full_location_warn = "'{0}' (full location: '{1}')".format(item['name'], full_location)

                    full_location_inner = inner_item.full_name.replace(': ', ' -> ')
                    full_location_inner_warn = "'{0}' (full location: '{1}')".format(inner_item['name'],
                                                                                     full_location_inner)

                    logger.warning('-SBN-: Overlapping field_locations found for Field {0} and Field {1}'
                                   .format(full_location_warn, full_location_inner_warn))

class TableObject(Object):
    """ Stores meta data about a single Table Object (i.e, Table_Character, Table_Binary or Table_Delimited) """

    def __init__(self, parent_filename, object_label, object_data=None):

        super(TableObject, self).__init__(parent_filename, object_label, object_data)

        self._set_meta_data(object_label)

        if 'name' in self._meta_data:
            self.id = self._meta_data['name']
        else:
            self.id = self._meta_data['local_identifier']

        if self.type != 'Table_Delimited':
            self._set_required_fixed_width()

    def _set_required_fixed_width(self):
        self._meta_data['records'] = int(self._meta_data['records'])
        self._meta_data['offset'] = int(self._meta_data['offset'])

        record_str = 'Record_' + self.type.split('_')[-1]
        self._meta_data[record_str]['record_length'] = int(self._meta_data[record_str]['record_length'])
        self.record = self._meta_data[record_str]

    def _set_meta_data(self, object_label):

        skip_fields = ['*Field_*', 'Uniformly_Sampled'] + self._meta_data.keys()
        optional_fields = xml_to_dict(object_label, skip_tags=skip_fields)

        for key, value in optional_fields.iteritems():
            self._meta_data[key] = value


class Field(OrderedDict):
    """ Stores meta data about a single <Field_*>, extended by all Field types """

    def __init__(self, *args, **kwds):
        super(Field, self).__init__(*args, **kwds)

        self.full_name = None
        self.group_repetition = 0
        self.group_level = 0

    def load(self, xml_field):
        self['name'] = xml_field.find('name').text
        self.full_name = self['name']

    def _set_optional_elements(self, xml_field):

        # Set scaling_factor and value_offset manually to allow casting
        if xml_field.find('scaling_factor') is not None:
            self['scaling_factor'] = cast_int_or_float(xml_field.find('scaling_factor').text)

        if xml_field.find('value_offset') is not None:
            self['value_offset'] = cast_int_or_float(xml_field.find('value_offset').text)

        # Set other optional fields
        optional_fields = xml_to_dict(xml_field, skip_tags=self.keys(), tag_modify=('field_', ''))

        for key, value in optional_fields.iteritems():
            self[key] = value


class FieldCharacter(Field):
    """ Stores meta data about a single <Field_Character> """

    def load(self, xml_field):

        super(FieldCharacter, self).load(xml_field)

        self['location'] = int(xml_field.find('field_location').text)
        self['length'] = int(xml_field.find('field_length').text)
        self['data_type'] = xml_field.find('data_type').text

        self._set_optional_elements(xml_field)


class FieldBinary(Field):
    """ Stores meta data about a single <Field_Binary> """

    def load(self, xml_field):

        super(FieldBinary, self).load(xml_field)

        self['location'] = int(xml_field.find('field_location').text)
        self['length'] = int(xml_field.find('field_length').text)
        self['data_type'] = xml_field.find('data_type').text

    #   self._set_packed_data_fields(xml_field)
        self._set_optional_elements(xml_field)

    def _set_packed_data_fields(self, xml_field):

        xml_packed_data = xml_field.find('Packed_Data_Fields')

        if xml_packed_data is not None:

            self['Packed_Data_Fields'] = OrderedDict()
            self['Packed_Data_Fields']['bit_fields'] = int(xml_packed_data.find('bit_fields').text)
            self['Packed_Data_Fields']['description'] = xml_packed_data.find('description').text
            self['Packed_Data_Fields']['Field_Bits'] = tuple()

            for xml_field_bit in xml_field.findall('Field_Bit'):

                field_bit = FieldBit()
                field_bit.load(xml_field_bit)

                self['Packed_Data_Fields']['Field_Bits'] += (field_bit, )


class FieldDelimited(Field):
    """ Stores meta data about a single <Field_Delimited> """

    def load(self, xml_field):

        super(FieldDelimited, self).load(xml_field)

        # Not yet implemented


class FieldBit(Field):
    """ Stores meta data about a single <Field_Bit> """

    def load(self, xml_field_bit):

        super(FieldBit, self).load(xml_field_bit)

        # Not yet implemented


class FieldUniformlySampled(Field):
    """ Stores meta data about a single <Uniformly_Sampled> """

    def load(self, xml_uniformly_sampled):

        self['name'] = xml_uniformly_sampled.find('sampling_parameter_name').text
        self.full_name = self['name']
        self['interval'] = cast_int_or_float(xml_uniformly_sampled.find('sampling_parameter_interval').text)
        self['first_value'] = cast_int_or_float(xml_uniformly_sampled.find('first_sampling_parameter_value').text)
        self['last_value'] = cast_int_or_float(xml_uniformly_sampled.find('last_sampling_parameter_value').text)

        self._set_optional_elements(xml_uniformly_sampled)

    def _set_optional_attributes(self, xml_field):

        # Set other optional fields
        optional_fields = xml_to_dict(xml_field, skip_tags=self.keys(), tag_modify=('sampling_parameter_', ''))

        for key, value in optional_fields.iteritems():
            self[key] = value


class Group(OrderedDict):
    """ Stores meta data about a single <Group_Field_*> """

    def __init__(self, *args, **kwds):
        super(Group, self).__init__(*args, **kwds)

        self.full_name = None
        self.cloned = False
        self.group_repetition = 0
        self.group_level = 0

    def load(self, xml_group_field):

        self['location'] = int(xml_group_field.find('group_location').text)
        self['length'] = int(xml_group_field.find('group_length').text)
        self['repetitions'] = int(xml_group_field.find('repetitions').text)
        self['fields'] = int(xml_group_field.find('fields').text)
        self['groups'] = int(xml_group_field.find('groups').text)

        self._set_optional_elements(xml_group_field)

    def _set_optional_elements(self, xml_group_field):

        optional_fields = xml_to_dict(xml_group_field, skip_tags=self.keys(), tag_modify=('group_', ''),
                                      skip_sub_tags=True)

        for key, value in optional_fields.iteritems():
            self[key] = value


# Returns true if table_element is a Group object (i.e. Group_Field_*), and false if it is not
def is_group(table_element):
    if isinstance(table_element, Group):
        return True

    return False
