import functools
import copy
import re
from math import floor
from Tkinter import Menu, Frame, Label,  Entry, END, Scrollbar

from pds4_viewer import DataObjectWindow
from pds4_logging import logger_init
from pds4_general_objects import Data
from pds4_table_objects import Field, is_group


# Safe import of numpy (not required)
try:
    import numpy as np
except ImportError:
    pass

# Initialize the logger
logger = logger_init()

#################################

class TableObjectWindow(DataObjectWindow):
    """ Window for displaying the Table_Character, Table_Binary and Table_Delimited data objects

    This window will display tabular data. After creating it, you should use the open_table() method to load the table
    that it needs to display.
    """

    def __init__(self, parent):

        # Create basic PDS4 Viewer
        super(TableObjectWindow, self).__init__(parent)

        # Add vertical scrollbar
        self._vert_scrollbar = Scrollbar(self._widget, orient='vertical', command=self._vertical_scroll)
        self._canvas.configure(yscrollcommand=self._set_vertical_scroll_length)
        self._vert_scrollbar.pack(side='right', fill='y')

        # Add horizontal scrollbar
        self._horz_scrollbar = Scrollbar(self._widget, orient='horizontal', command=self._horizontal_scroll)
        self._canvas.configure(xscrollcommand=self._set_horizontal_scrollbar_length)
        self._horz_scrollbar.pack(side='bottom', fill='x')

        # Pack the canvas and all its contents
        self._canvas.pack(side='left', expand=2, fill='both')

        # The window_resize binding adds/removes rows and columns onto the screen based on window size
        self._canvas.bind('<Configure>', self._window_resize)

        # Necessary for proper scrollbar resizing on window_resize
        self._canvas_frame.bind('<Configure>', lambda e: self._canvas.configure(scrollregion=self._canvas.bbox('all'),
                                                                                width=self._win_dimensions['width'],
                                                                                height=self._win_dimensions['height'])
                                )

        # These variables are used to store widgets, and info about them, used for displaying tabular data
        self._data_boxes = []
        self._background_boxes = []

    # Loads the table object, whose data as returned by read_table() is located at data.values()[table_index], into this
    # window and displays it for the user
    def open_table(self, table_object):

        self._append_to_title(" - Table '{0}'".format(table_object.id))

        # Get basic properties of the table being opened
        table_fields = get_fields_from_data(table_object.data)

        num_rows = table_object.meta_data['records']
        num_columns = len(table_fields)

        # Set properties of the table that was loaded, as well as display properties, for use
        self._data = {'table_fields': table_fields, 'display_data_type': 'formatted',
                      'num_rows': num_rows, 'num_columns': num_columns,
                      'num_display_rows': 0, 'display_start_row': 0,
                      'num_display_cols': 0, 'display_start_col': 0}

        # Create row count label on left
        box = Frame(self._canvas_frame, height=20, width=50)
        box.grid(row=0, column=0, pady=(10, 6), padx=14)

        label = Label(box, text='Row #', font='arial 8', bd=0)
        label.pack(fill='both', expand=2)

        # Adds new menu options used for manipulating the data
        self._add_data_menus()

        # Mark table as open, simulate window resizing to populate the window with the table's data
        self._data_open = True
        self._window_resize(None)

    # Changes the display_data_type and updates the on-screen values to use that display_data_type
    def _toggle_display_data_type(self, data_type):

        self._data['display_data_type'] = data_type
        self._update_vertical_table_display(self._vert_scrollbar.get())

    # Returns the data point located in the row and column indicies specified, formats it first if display_data_type is
    # set to formatted
    def _get_data_point(self, row, column):

        data = self._data['table_fields']
        data_point = data.values()[column][row]

        if self._data['display_data_type'] == 'formatted':
            meta_data = data.meta_data(data.keys()[column])

            try:
                if 'format' in meta_data:
                    data_point = _format_data_point(data_point, meta_data['format'])
            except ValueError:
                logger.warning("Unable to format value '{0}' into field_format '{1}'; displaying unformatted value"
                               .format(data_point, meta_data['format']))

        return data_point

    # Adds new menu options used for manipulating the data
    def _add_data_menus(self):

        # Adds an Export sub-menu to the File menu
        # file_menu = self._menu.winfo_children()[0]
        # export_menu = Menu(file_menu, tearoff=0)
        # file_menu.insert_cascade(0, label='Export', menu=export_menu)
        # export_menu.add_command(label='Export as txt table', command=OpenFile)

        # Add an Options menu
        options_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='Options', menu=options_menu)

        # Add a Data Formatting sub-menu to the Options menu
        formatting_menu = Menu(options_menu, tearoff=0)
        options_menu.add_cascade(label="Data formatting", menu=formatting_menu)

        data_type_formatted = functools.partial(self._toggle_display_data_type, 'formatted')
        data_type_unformatted = functools.partial(self._toggle_display_data_type, 'unformatted')
        formatting_menu.add_command(label='Show formatted data', command=data_type_formatted)
        formatting_menu.add_command(label='Show unformatted data', command=data_type_unformatted)

    # Draws the data from self.data on the screen by adding the appropriate widgets
    #
    # draw_row and draw_col are dictionaries with 'start' and 'stop' values, indicating the physical location of
    # the rows and columns that should be added to the screen. The exception being a row start value of -1, which
    # indicates that the row to add contains the 'column names' header (top row), and a column value of -1 which
    # indicates that the column to add contains the 'row count' (left most column). data_row_offset and data_col_offset
    # are used to indicate the difference between the physical location of the row/column in the table displayed on the
    # screen and the row/column of the data variable that should be used as to pull the value for that physical location
    # (e.g. get_data_point(row + data_row_offset, col + data_col_offset), where row is a value between draw_row's
    # 'start' and 'stop' values) and similar for col).
    def _draw_data(self, draw_row, draw_col, data_row_offset=0, data_col_offset=0):

        # Create column names
        if draw_row['start'] == -1:

            for column in range(draw_col['start'], draw_col['stop']):
                box = Frame(self._canvas_frame, height=20)
                box.grid(row=0, column=column + 1, padx=6, pady=(10, 4))

                interior_box = Frame(box, height=18, width=126)
                interior_box.pack()
                interior_box.pack_propagate(False)

                column_name_idx = column + data_col_offset

                entry = Entry(interior_box, borderwidth=0, font='arial 9 bold')
                entry.insert(0, self._data['table_fields'].keys()[column_name_idx])
                entry.configure(state='readonly')
                entry.pack(fill='both', expand=2)

                self._background_boxes.append({'box': box, 'row': -1, 'col': column})
                self._data_boxes.append({'entry': entry, 'row': -1, 'col': column})

        # Create row numbers
        if draw_col['start'] == -1:

            for row in range(draw_row['start'], draw_row['stop']):
                border_box = Frame(self._canvas_frame, height=20, width=50, background='black')
                border_box.grid(row=row + 1, column=0, padx=6, pady=2)

                interior_box = Frame(border_box, height=18, width=48, background='white')
                interior_box.pack(fill='both', expand=2, padx=1, pady=1)
                interior_box.pack_propagate(False)

                data_row = row + data_row_offset + 1

                entry = Entry(interior_box, borderwidth=0, readonlybackground='white', justify='center')
                entry.insert(0, data_row)
                entry.configure(state='readonly')
                entry.pack(fill='both', expand=2, padx=4)

                self._background_boxes.append({'box': border_box, 'row': row, 'col': -1})
                self._data_boxes.append({'entry': entry, 'row': row, 'col': -1})

        # Create values
        for column in range(draw_col['start'], draw_col['stop']):

            for row in range(draw_row['start'], draw_row['stop']):
                border_box = Frame(self._canvas_frame, height=20, width=130, background='black')
                border_box.grid(row=row + 1, column=column + 1, padx=4, pady=2)

                interior_box = Frame(border_box, height=18, width=128, background='white')
                interior_box.pack(fill='both', expand=2, padx=1, pady=1)
                interior_box.pack_propagate(False)

                data_row = row + data_row_offset
                data_column = column + data_col_offset

                entry = Entry(interior_box, borderwidth=0, readonlybackground='white')
                entry.insert(0, self._get_data_point(data_row, data_column))
                entry.configure(state='readonly')
                entry.pack(fill='both', expand=2, padx=4)

                self._background_boxes.append({'box': border_box, 'row': row, 'col': column})
                self._data_boxes.append({'entry': entry, 'row': row, 'col': column})

    # Erases the appropriate widgets from the screen
    #
    # erase_row and erase_col are dictionaries with 'start' and 'stop' values, indicating the physical location of
    # the rows and columns that should be removed from the table being displayed on the screen
    def _erase_data(self, erase_row, erase_col):

        remove_bg_box_idxs = [i for i, data_box in enumerate(self._data_boxes)
                              if erase_row['start'] <= data_box['row'] <= erase_row['stop']
                              if erase_col['start'] <= data_box['col'] <= erase_col['stop']]

        self._data_boxes = [data_box for i, data_box in enumerate(self._data_boxes) if i not in remove_bg_box_idxs]
        deleted_boxes = []

        # This should use destroy(). However, there appears to be a TKInter bug in at least Windows 7 that will
        # resize the window back up to its previous size under most conditions if destroy() is used. As a
        # workaround we use grid_forget(), allowing a small memory leak only while the viewer is open and if its resized
        for i in range(0, len(remove_bg_box_idxs)):
            box = self._background_boxes[remove_bg_box_idxs[i]]
            deleted_boxes.append(box)
            box['box'].grid_forget()

        for box in deleted_boxes:
            self._background_boxes.remove(box)

    # Called when the vertical scrollbar is used, this method adjusts the scrollbar's position to the new position and
    # also calls _update_vertical_table_display()
    def _vertical_scroll(self, action, first_cmd, second_cmd=''):

        position = self._vert_scrollbar.get()
        scrollbar_length = position[1] - position[0]

        if second_cmd == '':
            # This means scrollbar was dragged; first_cmd = offset
            self._vert_scrollbar.set(float(first_cmd), float(first_cmd) + scrollbar_length)
        else:
            # This means scrollbar scroll button clicked; first_cmd = +1 or -1; second_cmd = 'scroll what'
            step = scrollbar_length / 5

            if int(first_cmd) > 0:
                # Scroll down

                if (position[1] + step) > 1:
                    self._vert_scrollbar.set(1 - scrollbar_length, 1)
                else:
                    self._vert_scrollbar.set(position[0] + step, position[1] + step)

            else:
                # Scroll up

                if (position[0] - step) < 0:
                    self._vert_scrollbar.set(0, scrollbar_length)
                else:
                    self._vert_scrollbar.set(position[0] - step, position[1] - step)

        self._update_vertical_table_display(self._vert_scrollbar.get())

    # Called when the horizontal scrollbar is used, this method adjusts the scrollbar's position to the new position and
    # also calls _update_horizontal_table_display()
    def _horizontal_scroll(self, action, first_cmd, second_cmd=''):
        position = self._horz_scrollbar.get()
        scrollbar_length = position[1] - position[0]

        if second_cmd == '':
            # This means scrollbar was dragged; first_cmd = offset
            self._horz_scrollbar.set(float(first_cmd), float(first_cmd) + scrollbar_length)
        else:
            # This means scrollbar scroll button clicked; first_cmd = +1 or -1; second_cmd = 'scroll what'
            step = scrollbar_length / 5

            if int(first_cmd) > 0:
                # Scroll right

                if (position[1] + step) > 1:
                    self._horz_scrollbar.set(1 - scrollbar_length, 1)
                else:
                    self._horz_scrollbar.set(position[0] + step, position[1] + step)

            else:
                # Scroll left

                if (position[0] - step) < 0:
                    self._horz_scrollbar.set(0, scrollbar_length)
                else:
                    self._horz_scrollbar.set(position[0] - step, position[1] - step)

        self._update_horizontal_table_display(self._horz_scrollbar.get())

    # Adjusts the length of the vertical scrollbar depending on number of rows in the table and on window resize
    def _set_vertical_scroll_length(self, start_offset, end_offset):

        if self._data_open:
            position = self._vert_scrollbar.get()

            start_offset = position[0]
            end_offset = position[0] + float(self._data['num_display_rows']) / self._data['num_rows']

            if end_offset > 1:
                start_offset -= end_offset - 1
                end_offset = 1

            self._vert_scrollbar.set(start_offset, end_offset)
            self._widget.update_idletasks()
        else:
            self._vert_scrollbar.set(start_offset, end_offset)

    # Adjusts the length of the horizontal scrollbar depending on number of columns in the table and on window resize
    def _set_horizontal_scrollbar_length(self, start_offset, end_offset):

        if self._data_open:
            position = self._horz_scrollbar.get()

            start_offset = position[0]
            end_offset = position[0] + float(self._data['num_display_cols']) / self._data['num_columns']

            if end_offset > 1:
                start_offset -= end_offset - 1
                end_offset = 1

            self._horz_scrollbar.set(start_offset, end_offset)
            self._widget.update_idletasks()
        else:
            self._horz_scrollbar.set(start_offset, end_offset)

    # Adjusts data values of self._data_boxes (values in boxes being displayed) to match the vertical position the user
    # is currently viewing
    def _update_vertical_table_display(self, position):

        # Get the new display_start_row of the table (once it is updated)
        start_row = int(floor(self._data['num_rows'] * position[0]))

        if (start_row + self._data['num_display_rows']) > self._data['num_rows']:
            start_row = self._data['num_rows'] - self._data['num_display_rows']

        # Loop over each data box
        for data_box in self._data_boxes:

            # Skip the 'column name' row since it is statically displayed at the top regardless of scrolling, otherwise
            # adjust the entry contents
            if data_box['row'] != -1:

                data_row_idx = data_box['row'] + start_row
                data_col_idx = data_box['col'] + self._data['display_start_col']

                data_box['entry'].configure(state='normal')
                data_box['entry'].delete(0, 'end')

                # First column is the 'row count' row
                if data_box['col'] == -1:
                    data_box['entry'].insert(0, data_row_idx + 1)

                # All other columns get data from the data table
                else:
                    data_box['entry'].insert(0, self._get_data_point(data_row_idx, data_col_idx))

                data_box['entry'].configure(state='readonly')

        self._data['display_start_row'] = start_row

    # Adjusts data values of self._data_boxes (values in boxes being displayed) to match the horizontal position the
    # user is currently viewing
    def _update_horizontal_table_display(self, position):

        # Get the new display_start_col of the table (once it is updated)
        start_col = int(floor(self._data['num_columns'] * position[0]))

        if (start_col + self._data['num_display_cols']) > self._data['num_columns']:
            start_col = self._data['num_columns'] - self._data['num_display_cols']

        # Loop over each entry
        for data_box in self._data_boxes:

            # Skip the 'row count' column since it is statically displayed at the left regardless of scrolling,
            # otherwise adjust the entry contents
            if data_box['col'] != -1:

                data_row_idx = data_box['row'] + self._data['display_start_row']
                data_col_idx = data_box['col'] + start_col

                data_box['entry'].configure(state='normal')
                data_box['entry'].delete(0, END)

                # First row is the 'column names' header row
                if data_box['row'] == -1:
                    data_box['entry'].insert(0, self._data['table_fields'].keys()[data_col_idx])

                # All other rows get data from the data table
                else:
                    data_box['entry'].insert(0, self._get_data_point(data_row_idx, data_col_idx))

                data_box['entry'].configure(state='readonly')

        self._data['display_start_col'] = start_col

    # Called when the window has been resized, this method calculates which rows and columns are able to fit with-in the
    # new window size and then calls either _draw_data() or _erase_data() to add or remove the appropriate rows/columns
    def _window_resize(self, event):

        self._update_window_dimensions()

        if not self._data_open:
            return

        # Determine the number of rows and columns that will fit in the viewer (with current window size).
        # Rows follows the formula: (Window_Height - Titlebar_Height - Header_Height) / Row_Height
        # Columns follows the formula: (Window_Width - Scrollbar__and_Frame_Height - Row_Count_Width) / Column_Width
        num_display_rows = int(floor((self._win_dimensions['height'] - 20 - 30) / 24.))
        num_display_cols = int(floor((self._win_dimensions['width'] - 50 - 56) / 134.))

        if num_display_rows > self._data['num_rows']:
            num_display_rows = self._data['num_rows']
        elif num_display_rows < 0:
            num_display_rows = 0

        if num_display_cols > self._data['num_columns']:
            num_display_cols = self._data['num_columns']
        elif num_display_cols < 0:
            num_display_cols = 0

        num_missing_rows = num_display_rows - self._data['num_display_rows']
        num_missing_cols = num_display_cols - self._data['num_display_cols']

        # Calculate start and stop rows/columns for new rows/columns to add
        if num_missing_rows > 0 or num_missing_cols > 0:

            # Row and column offsets which indicate the difference in position of the row or column being added and the
            # position of the value in the data. I.e., a row being added would have the value:
            # value = get_data_point(draw_row['start'] + data_row_offset)
            data_row_offset = self._data['display_start_row']
            data_column_offset = self._data['display_start_col']

            # If window has been vertically enlarged, add new rows to fit
            if num_missing_rows > 0:

                draw_row = {'start': self._data['num_display_rows'],
                            'stop': self._data['num_display_rows'] + num_missing_rows}

                draw_col = {'start': 0, 'stop': self._data['num_display_cols']}

                # Special case where we need to add rows at the beginning because user has tried to expand screen after
                # scrolling all the way down
                if data_row_offset + draw_row['stop'] > self._data['num_rows']:

                    # Erases all rows, except header row, from the screen
                    erase_row = {'start': 0,
                                 'stop': self._data['num_display_rows']}

                    self._erase_data(erase_row, draw_col)

                    # We start from beginning and add all requested rows
                    draw_row = {'start': 0,
                                'stop': num_display_rows}
                    data_row_offset = self._data['num_rows'] - num_display_rows

                    # Adjust the display start row to account for the new starting value
                    self._data['display_start_row'] = data_row_offset

                # Draw additional data rows
                self._data['num_display_rows'] = num_display_rows
                self._draw_data(draw_row, draw_col, data_row_offset, data_column_offset)

                # Draw additional 'row count' rows if necessary
                row_count_drawn = next((True for data_box in self._data_boxes
                                        if data_box['col'] == -1
                                        if data_box['row'] >= draw_row['stop'] - 1
                                        ), False)
                if not row_count_drawn:
                    self._draw_data(draw_row, {'start': -1, 'stop': -1}, data_row_offset, 0)

            # If window has been horizontally enlarged, add new columns to fit
            if num_missing_cols > 0:

                draw_col = {'start': self._data['num_display_cols'],
                            'stop': self._data['num_display_cols'] + num_missing_cols}

                draw_row = {'start': 0, 'stop': self._data['num_display_rows']}

                # Special case where we need to add columns at the beginning because user has tried to expand screen
                # after scrolling all the way to the right
                if data_column_offset + draw_col['stop'] > self._data['num_columns']:

                    # Erases all columns, except row count column, from the screen
                    erase_col = {'start': 0,
                                 'stop': self._data['num_display_cols']}

                    erase_row = {'start': -1,
                                 'stop': self._data['num_display_rows']}

                    self._erase_data(erase_row, erase_col)

                    # We start from beginning and add all requested columns
                    draw_col = {'start': 0,
                                'stop': num_display_cols}
                    data_column_offset = self._data['num_columns'] - num_display_cols

                    # Adjust the display start column to account for the new starting value
                    self._data['display_start_col'] = data_column_offset

                # Draw additional data columns
                self._data['num_display_cols'] = num_display_cols
                self._draw_data(draw_row, draw_col, data_row_offset, data_column_offset)

                # Draw additional 'column names' row if necessary
                column_names_drawn = next((True for data_box in self._data_boxes
                                           if data_box['row'] == -1
                                           if data_box['col'] >= draw_col['stop'] - 1
                                           ), False)
                if not column_names_drawn:
                    self._draw_data({'start': -1, 'stop': -1}, draw_col, 0, data_column_offset)

        elif num_missing_rows < 0 or num_missing_cols < 0:

            # Window has been vertically shrunk, erase extra rows that are outside the screen
            if num_missing_rows < 0:
                draw_row = {'start': self._data['num_display_rows'] + num_missing_rows,
                            'stop': self._data['num_display_rows']}

                draw_col = {'start': -1, 'stop': self._data['num_display_cols']}

                # Remove extra rows
                self._data['num_display_rows'] = num_display_rows
                self._erase_data(draw_row, draw_col)

            # Window has been horizontally shrunk, erase extra columns that are outside the screen
            if num_missing_cols < 0:
                draw_col = {'start': self._data['num_display_cols'] + num_missing_cols,
                            'stop': self._data['num_display_cols']}

                draw_row = {'start': -1, 'stop': self._data['num_display_rows']}

                # Remove extra columns
                self._data['num_display_cols'] = num_display_cols
                self._erase_data(draw_row, draw_col)


# Formats a single data point according to field_format
def _format_data_point(datum, field_format):
    python_field_format = '0:'

    # The left justify character for python is '<' instead of '-' as in the PDS4 standard
    if field_format[1] == '-':
        python_field_format += '<'
        field_format = field_format[1:]

    # Fix for precision/specifier in case of integer formats; python does not support precision for integers
    int_width = None
    int_specifier = None

    if 'd' in field_format or 'o' in field_format:

        specifier_regex = re.compile('([0-9]+)\.([0-9]+)')
        specifier_search = specifier_regex.search(field_format)

        if specifier_search is not None:
            int_width = int(specifier_search.group(1))
            int_specifier = int(specifier_search.group(2))
            python_field_format += '0'
            field_format = re.sub(specifier_regex, str(int_specifier), field_format)

    # Format datum according to field_format (for the most part the PDS4 standard format is the same as Python)
    python_field_format = '{' + python_field_format + field_format[1:] + '}'
    formatted_datum = python_field_format.format(datum)

    # Finish up specifier fix in case of integers (discussed above) by adding any necessary leading spaces
    if int_width is not None and int_specifier is not None:

        # Ensures that too short a specifier does not replace actual data with zeros (i.e., padding only)
        if int_specifier < len(formatted_datum):
            int_specifier = len(formatted_datum)

        int_format = '{0:>' + str(int_width) + '.' + str(int_specifier) + '}'
        formatted_datum = int_format.format(formatted_datum)

    return formatted_datum


def get_fields_from_data(table_data):
    table_fields = Data()

    # Loop over all items in table_data
    for key, item in copy.deepcopy(table_data).iteritems():

        item_meta_data = table_data.meta_data(key)

        if hasattr(item[0], '__len__') and (not isinstance(item[0], str)):

            if is_group(item_meta_data):

                for single_repetition in item:

                    table_fields_recurse = get_fields_from_data(single_repetition)

                    for recurse_key, recurse_value in table_fields_recurse.iteritems():
                        table_fields.set(recurse_key, recurse_value, table_fields_recurse.meta_data(recurse_key))

            elif is_group(item_meta_data[0]):

                for i, single_group in enumerate(item):

                    for single_repetition in single_group:

                        table_fields_recurse = get_fields_from_data(single_repetition)

                        for recurse_key, recurse_value in table_fields_recurse.iteritems():
                            table_fields.set(recurse_key, recurse_value, table_fields_recurse.meta_data(recurse_key))

            else:

                for i, single_field in enumerate(item):

                    field_metadata = item_meta_data[i]
                    table_fields.set(field_metadata.full_name, single_field, field_metadata)

        elif isinstance(item_meta_data, Field):
            table_fields.set(item_meta_data.full_name, item, item_meta_data)

    return table_fields
