import os
import functools
import traceback
from math import floor
import xml.etree.ElementTree as ET
from Tkinter import Tk, Toplevel, Scrollbar, Text, Image, Menu, Frame, Label, Button, Canvas, Entry
from tkFileDialog import askopenfilename

import pds4_read
from pds4_logging import logger_init

# Safe import of numpy
try:
    import numpy as np
except ImportError:
    pass

# Initialize the logger
logger = logger_init()

#################################


class PDS4Viewer(object):
    """ Main class and window manager for PDS4 Viewer

    PDS4 Viewer hides the root window, and uses additional Toplevel Windows to display its contents. Once all PDS4
    Viewer open Windows are the closed the root window is then also closed automatically. Additionally the root and all
    other Windows can be closed anytime via .quit(). PDS4 Viewer will auto-remove any Window that is physically closed
    by the user.
    """

    def __init__(self, root):

        self._root = root
        self._root.withdraw()

        self._open_windows = []
        self.last_open_dir = None

    # Adds window to PDS4 Viewer
    def add_window(self, window):

        self._open_windows.append(window)

        return len(self._open_windows)

    # Removes window from PDS4 Viewer
    def remove_window(self, window):

        self._open_windows = [open_window for open_window in self._open_windows if open_window.id != window.id]

        if len(self._open_windows) == 0:
            self.quit()

    # Quits PDS4 Viewer, closing all open windows
    def quit(self):

        for window in self._open_windows:
            self.remove_window(window)

        if self._root is not None:
            self._root.destroy()
            self._root = None

class Window(object):
    """ Base class of any window for PDS4 Viewer """

    def __init__(self, viewer):

        self._widget = Toplevel()
        self._widget.title('PDS4 Viewer')

        self._viewer = viewer
        self._set_logo()

        self._id = self._viewer.add_window(self)
        self._widget.protocol('WM_DELETE_WINDOW', self.close)

        self._win_dimensions = {'height': 0, 'width': 0}
        self._update_window_dimensions()

    def close(self):

        self._widget.destroy()
        self._viewer.remove_window(self)

    @property
    def id(self):
        return self._id

    def _center_window(self):
        self._update_window_dimensions()
        x = (self._widget.winfo_screenwidth() // 2) - (self._win_dimensions['width'] // 2)
        y = (self._widget.winfo_screenheight() // 2) - (self._win_dimensions['height'] // 2)
        self._widget.geometry(
            '{0}x{1}+{2}+{3}'.format(self._win_dimensions['width'], self._win_dimensions['height'], x, y))

    def _fit_to_content(self):
        self._widget.winfo_toplevel().wm_geometry('')
        self._widget.update_idletasks()
        self._widget.geometry('{0}x{1}'.format(self._widget.winfo_width(), self._widget.winfo_height()))
        self._update_window_dimensions()

    def _append_to_title(self, subtitle):
        self._widget.title(self._widget.title() + subtitle)

    def _set_logo(self):
        logo_path = os.path.join(os.path.dirname(__file__), 'logo.gif')

        if os.access(logo_path, os.R_OK):
            logo = Image('photo', file=logo_path)
            self._widget.tk.call('wm', 'iconphoto', self._widget._w, logo)

    def _update_window_dimensions(self):
        self._widget.update_idletasks()
        self._win_dimensions = {'height': self._widget.winfo_height(), 'width': self._widget.winfo_width()}


class ObjectListWindow(Window):
    """ Window that summarizes the objects, showing some of their properties and giving buttons to open them """

    def __init__(self, viewer):

        # Set initial necessary variables and do other required initialization procedures
        super(ObjectListWindow, self).__init__(viewer)

        # Create menu
        self._menu = Menu(self._widget)
        self._widget.config(menu=self._menu)

        file_menu = Menu(self._widget, tearoff=0)
        file_menu.add_command(label='Open...', command=self._open_file_box)
        file_menu.add_separator()
        file_menu.add_command(label='Exit', command=self._viewer.quit)
        self._menu.add_cascade(label='File', menu=file_menu)

        # Initialize object list window variables
        self._last_open_dir = None
        self._label_open = False

    # Opens the label, reads in any objects it contains, calls _draw_summary()
    def open_label(self, filename, from_existing_objects=None, quiet=False, object_num=None, object_name=None,
                   object_lid=None):

        # Determine whether numpy was imported
        use_numpy = False
        if np:
            use_numpy = True

        # Open the label and data, or show an error if one occurs
        try:
            if from_existing_objects is None:
                object_list = pds4_read.pds4_read(filename, quiet=quiet, use_numpy=use_numpy,
                                                  object_num=object_num, object_name=object_name, object_lid=object_lid)

                self._set_last_open_dir(filename)
            else:
                object_list = from_existing_objects

            self._draw_summary(object_list)

            self._label_open = True
            self._append_to_title(" - Data Object Summary for '{0}'".format(filename))

        except Exception:

            ScrollTextWindow(self._viewer, 'An Error Occurred!', traceback.format_exc())
            self.close()
            raise

    # Draws a summary of the opened label onto the screen
    def _draw_summary(self, object_list):

        # Add Info menu
        info_menu = Menu(self._menu, tearoff=0)
        self._menu.add_cascade(label='Info', menu=info_menu)

        info_menu.add_command(label='Show read-in log',
                    command=lambda: ScrollTextWindow(self._viewer, 'Label/Data Read-in Log', object_list.read_in_log))

        # Create canvas and the header of the summary
        canvas = Canvas(self._widget)
        canvas.pack(fill='both', expand=1)

        index = Label(canvas, text='Index', font='arial 10 bold')
        index.grid(row=0, column=0, padx=(25, 0), pady=(2, 7), sticky='w')

        name = Label(canvas, text='Name', font='arial 10 bold')
        name.grid(row=0, column=1, padx=(80, 105), pady=(2, 7), sticky='w')

        type = Label(canvas, text='Type', font='arial 10 bold')
        type.grid(row=0, column=2, pady=(2, 7), sticky='w')

        view = Label(canvas, text='View', font='arial 10 bold')
        view.grid(row=0, column=3, padx=(105, 55), pady=(2, 7), sticky='w')

        # Create canvas frame and show all object meta data inside it
        canvas_frame = Frame(canvas, takefocus=False, width=600, height=31 * len(object_list))
        canvas.create_window((0, 0), window=canvas_frame, anchor='nw')
        canvas_frame.grid(row=1, column=0, columnspan=4, sticky='w')

        for i, object in enumerate(object_list):

            index = Label(canvas_frame, text=i + 1, font='arial 9')
            index_padx = len(str(i)) * 3
            index.place(x=43 - index_padx, y=30 * i)

            name_text = object.id.strip()
            if len(name_text) >= 30:
                name_text = name_text[:29] + '...'

            name = Label(canvas_frame, text=name_text, font='arial 9')
            name_padx = floor(len(name_text) * 2.8)
            name.place(x=43 + 140 - name_padx, y=30 * i)

            type = Label(canvas_frame, text=object.type, font='arial 9')
            type_padx = floor(len(object.type) * 2.8)
            type.place(x=43 + 140 + 160 - type_padx, y=30 * i)

            open_label = functools.partial(self._open_label, object_list.label)
            view_label = Button(canvas_frame, text='   Label   ', font='arial 9 bold', command=open_label)
            view_label.place(x=43 + 140 + 160 + 105, y=30 * i)

            open_table = functools.partial(self._open_table, object)
            view_table = Button(canvas_frame, text='   Table   ', font='arial 9 bold', command=open_table)
            view_table.place(x=43 + 140 + 160 + 105 + 60, y=30 * i)

        # Set window to fit content and add line diving header and summary data
        self._fit_to_content()
        canvas.create_line(5, 25, self._widget.winfo_width() - 5, 25)

    def _open_label(self, label):

        label_window = LabelWindow(self._viewer, ET.tostring(label))

    def _open_table(self, table_object):

        from pds4_table_viewer import TableObjectWindow

        table_window = TableObjectWindow(self._viewer)
        table_window.open_table(table_object)

    def _open_file_box(self):

        filetypes = [('All Files', '*'), ('XML Files', 'xml')]

        if self._viewer.last_open_dir is None:
            filename = askopenfilename(filetypes=filetypes)
        else:
            filename = askopenfilename(filetypes=filetypes, initialdir=self._viewer.last_open_dir)

        if filename.strip() != '':

            if self._label_open:
                summary_window = ObjectListWindow(self._viewer)
                summary_window.open_label(filename)
            else:
                self.open_label(filename)

    def _set_last_open_dir(self, filename):
        self._viewer.last_open_dir = os.path.dirname(filename)


class DataObjectWindow(Window):
    """ Window for displaying a single data object; specific data objects (e.g. Table_Character) extend this class """

    def __init__(self, viewer):
        # Set initial necessary variables and do other required initialization procedures
        super(DataObjectWindow, self).__init__(viewer)

        # Create File Menu
        self._menu = Menu(self._widget)
        self._widget.config(menu=self._menu)

        file_menu = Menu(self._menu, tearoff=0)
        file_menu.add_command(label='Close', command=self.close)
        file_menu.add_command(label='Close All', command=self._viewer.quit)
        self._menu.add_cascade(label='File', menu=file_menu)

        # Initialize data variables
        self._data = {}
        self._data_open = False

        # Initialize window size
        width = self._widget.winfo_screenwidth() / 2
        height = self._widget.winfo_screenheight() / 2
        self._widget.geometry('{0}x{1}'.format(width, height))
        self._center_window()

        # Create basic window structure and variables
        self._canvas = Canvas(self._widget, takefocus=False, bd=0, highlightthickness=0)
        self._canvas_frame = Frame(self._canvas, takefocus=False)
        self._canvas.create_window((0, 0), window=self._canvas_frame, anchor='nw')
        self._canvas_frame.pack(side='left', anchor='nw')

        self._vert_scrollbar = None
        self._horz_scrollbar = None


class ScrollTextWindow(Window):
    """ Window used to display messages inside a scrollable text area """

    def __init__(self, viewer, message_header, message):

        # Set initial necessary variables and do other required initialization procedures
        super(ScrollTextWindow, self).__init__(viewer)

        # Add header
        header = Label(self._widget, text=message_header, font='arial 15 bold')
        header.pack(pady=10, side='top')

        # Add buttons Ok and Copy buttons below the text pad first so they remain during resizing
        button_frame = Frame(self._widget)
        button_frame.pack(pady=10, side='bottom')

        ok_button = Button(button_frame, text='  Ok  ', font='arial 9 bold', command=self.close)
        ok_button.pack(side='left', padx=(0, 5))

        copy_button = Button(button_frame, text='  Copy  ', font='arial 9 bold', command=self._copy_message)
        copy_button.pack(side='left')

        # Add text pad
        self._text_pad = None
        self._draw_text_pad()
        self._text_pad.insert('end', message)
        self._text_pad.config(state='disabled')
        self._text_pad.see('end')

        self._fit_to_content()
        self._center_window()

    def _draw_text_pad(self):
        frame = Frame(self._widget)
        self._text_pad = Text(frame, wrap='none', width=50, height=20)

        vert_scrollbar = Scrollbar(frame, orient='vertical', command=self._text_pad.yview)
        self._text_pad.configure(yscrollcommand=vert_scrollbar.set)

        horz_scrollbar = Scrollbar(frame, orient='horizontal', command=self._text_pad.xview)
        self._text_pad.configure(xscrollcommand=horz_scrollbar.set)

        vert_scrollbar.pack(side='right', fill='y')
        horz_scrollbar.pack(side='bottom', fill='x')
        self._text_pad.pack(side='left', expand=2, fill='both')
        frame.pack(expand=2, fill='both')

    def _copy_message(self):
        message = self._text_pad.get("1.0", 'end-1c')

        self._widget.clipboard_clear()
        self._widget.clipboard_append(message)

class LabelWindow(Window):
    """ Window used to display the XML content of a label """

    def __init__(self, viewer, string_label):

        # Set initial necessary variables and do other required initialization procedures
        super(LabelWindow, self).__init__(viewer)

        # Add header
        header = Label(self._widget, text='Label', font='arial 15 bold')
        header.pack(pady=10, side='top')

        # Add search buttons below text pad first so they remain during resizing
        search_frame = Frame(self._widget)

        search_entry = Entry(search_frame, borderwidth=0, font='arial 9 bold')
        search_entry.pack(side='left', padx=(0, 5))

        # Add text pad
        self._text_pad = None
        self._draw_text_pad()
        self._text_pad.see('0.0')
        self._text_pad.insert('1.0', string_label)
        self._text_pad.config(state='disabled')

        self._fit_to_content()
        self._center_window()

    def _draw_text_pad(self):

        frame = Frame(self._widget)
        self._text_pad = Text(frame, wrap='none', width=100, height=20)

        vert_scrollbar = Scrollbar(frame, orient='vertical', command=self._text_pad.yview)
        self._text_pad.configure(yscrollcommand=vert_scrollbar.set)

        horz_scrollbar = Scrollbar(frame, orient='horizontal', command=self._text_pad.xview)
        self._text_pad.configure(xscrollcommand=horz_scrollbar.set)

        vert_scrollbar.pack(side='right', fill='y')
        horz_scrollbar.pack(side='bottom', fill='x')
        self._text_pad.pack(side='left', expand=2, fill='both')
        frame.pack(expand=2, fill='both')

def pds4_viewer(filename=None, from_existing_objects=None, quiet=False, object_num=None, object_name=None,
                object_lid=None):

    root = Tk()
    viewer = PDS4Viewer(root)

    summary_window = ObjectListWindow(viewer)
    if filename is not None:
        summary_window.open_label(filename, from_existing_objects=from_existing_objects, quiet=quiet,
                                  object_num=object_num, object_name=object_name, object_lid=object_lid)

    root.mainloop()


if __name__ == '__main__':

    # Safe import of argparse
    try:
        import argparse
    except ImportError:
        import pds4_tools.argparse_26 as argparse

    # Create program arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("filename", help="Filename, including full path, of the label", nargs='?', default=None)
    parser.add_argument("--quiet", help="Suppresses all info/warnings", action="store_true", default=False)
    parser.add_argument("--object_num", help="Only reads the data object specified by zero-based order (integer)")
    parser.add_argument("--object_name", help="Only reads the data object specified by name")
    parser.add_argument("--object_lid", help="Only reads the data object specified by local identifier")

    args = parser.parse_args()

    pds4_viewer(args.filename, object_num=args.object_num, object_name=args.object_name,
                object_lid=args.object_lid)
